#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Core datastructures
===================

Core datastructure for PATATO. This module provides structures such as Reconstruction, UnmixedData etc., providing
commonly-used functions.
"""
