#  Copyright (c) Thomas Else 2023.
#  License: MIT

"""
Time-series data processing module
=========================================

This module provides functions for commonly-used pre-processing procedures used in photoacoustic imaging.
This includes, for example, band-pass filtering, Hilbert transforms and interpolation.
"""